#
# Copyright 2019 Stephan Müller
#
# Licensed under the MIT license

"""Module ``pylint_gitlab``."""

from pylint_gitlab.reporter import GitlabCodeClimateReporter, GitlabPagesHtmlReporter


def register(linter):
    """Register the reporter classes with the linter."""
    linter.register_reporter(GitlabCodeClimateReporter)
    linter.register_reporter(GitlabPagesHtmlReporter)



#
# Copyright 2019 Stephan Müller
#
# Licensed under the MIT license

"""Tests for ``pylint_gitlab.reporter``."""

import json
import os
from io import StringIO

import pytest
from pylint.lint import PyLinter


@pytest.mark.parametrize("use_load_plugins", [False, True])
def test_gitlab_pages_html_reporter(use_load_plugins):
    """Tests for ``pylint_gitlab.reporter.GitlabPagesHtmlReporter()``."""

    if use_load_plugins:
        plugins = ["pylint_gitlab"]
        reporter = "gitlab-pages-html"
    else:
        plugins = []
        reporter = "pylint_gitlab.GitlabPagesHtmlReporter"

    output = StringIO()
    linter = PyLinter()

    linter.load_plugin_modules(plugins)
    linter.set_option("output-format", reporter)
    linter.set_option("persistent", False)
    linter.load_default_plugins()

    reporter = linter.reporter
    reporter.set_output(output)
    reporter.CI_PROJECT_URL = "https://example.org"
    reporter.CI_COMMIT_REF_NAME = "branch"

    linter.open()

    linter.set_current_module("b")
    linter.add_message("line-too-long", line=2, args=(1, 2))
    linter.add_message("line-too-long", line=1, args=(1, 2))

    linter.set_current_module("a")
    linter.add_message("line-too-long", line=1, args=(1, 2))

    # we call this method because we didn't actually run the checkers
    reporter.display_messages(None)

    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "report.html"), "r") as file:
        expected_result = file.read()
    assert output.getvalue() == expected_result


@pytest.mark.parametrize("use_load_plugins", [False, True])
def test_gitlab_code_climate_reporter(use_load_plugins):
    """Tests for ``pylint_gitlab.reporter.GitlabCodeClimateReporter()``."""

    if use_load_plugins:
        plugins = ["pylint_gitlab"]
        reporter = "gitlab-codeclimate"
    else:
        plugins = []
        reporter = "pylint_gitlab.GitlabCodeClimateReporter"

    output = StringIO()
    linter = PyLinter()

    linter.load_plugin_modules(plugins)
    linter.set_option("output-format", reporter)
    linter.set_option("persistent", False)
    linter.load_default_plugins()

    reporter = linter.reporter
    reporter.set_output(output)

    linter.open()

    linter.set_current_module("0123")
    linter.add_message("line-too-long", line=1, args=(1, 2))

    # we call this method because we didn't actually run the checkers
    reporter.display_messages(None)

    expected_result = [{
        "description": "C0301: Line too long (1/2)",
        "severity": "minor",
        "location": {
            "path": "0123",
            "lines": {
                "begin": 1,
            }
        },
        "fingerprint": "21e7b7887a841a99f73b9cf1771bf387"
    }]
    report_result = json.loads(output.getvalue())
    assert report_result == expected_result



#
# Copyright 2019 Stephan Müller
#
# Licensed under the MIT license

"""Gitlab codequality reporter"""
from __future__ import absolute_import, print_function

import hashlib
import html
import json
import os
import sys
from collections import OrderedDict
from itertools import groupby

from jinja2 import FileSystemLoader, Environment
from pylint.interfaces import IReporter
from pylint.reporters import BaseReporter


class GitlabPagesHtmlReporter(BaseReporter):
    """Report messages in HTML document with links to GitLab source code."""

    __implements__ = IReporter
    name = "gitlab-pages-html"
    extension = "html"

    CI_PROJECT_URL = os.getenv("CI_PROJECT_URL", "")
    CI_COMMIT_REF_NAME = os.getenv("CI_COMMIT_REF_NAME", "")

    _MSG_TYPES = {
        "info": "table-primary",
        "convention": "table-primary",
        "refactor": "table-primary",
        "warning": "table-warning",
        "error": "table-error",
        "fatal": "table-error",
    }

    _COLUMN_NAMES = OrderedDict({
        "line": "Line",
        "column": "Column",
        "type": "Type",
        "obj": "Object",
        "message": "Message",
        "symbol": "Symbol",
        "message_id": "Message Id",
    })

    def __init__(self, output=sys.stdout):
        BaseReporter.__init__(self, output)
        self.messages = []

    def handle_message(self, msg):
        """Manage message of different type and in the context of path."""
        self.messages.append(msg)

    def display_messages(self, layout):
        """Launch layouts display."""

        ordered_messages = OrderedDict()

        sorted_messages = sorted(self.messages, key=lambda msg: msg.module)
        for module, messages in groupby(sorted_messages, lambda msg: msg.module):
            ordered_messages.update({module: list()})
            for message in list(messages):
                ordered_messages[module].append({
                    "class": self._MSG_TYPES[message.category],
                    "url": self.CI_PROJECT_URL + "/blob/" + self.CI_COMMIT_REF_NAME + "/" + message.path + "#L" + str(
                        message.line),
                    "path": message.path,
                    "row": OrderedDict({
                        "line": message.line,
                        "column": message.column,
                        "type": message.category,
                        "obj": message.obj,
                        "message": message.msg.replace("\n", "<br />"),
                        "symbol": message.symbol,
                        "message_id": message.msg_id,
                    })
                })

        template = Environment(
            loader=FileSystemLoader(
                searchpath=os.path.dirname(os.path.abspath(__file__))
            )
        ).get_template("templates/report.html.j2")

        print(template.render(data=ordered_messages, column_names=self._COLUMN_NAMES), file=self.out)

    def display_reports(self, layout):
        """Do nothing."""

    def _display(self, layout):
        """Do nothing."""


class GitlabCodeClimateReporter(BaseReporter):
    """Report messages and layouts in Gitlab Code Climate format. Read more on
    https://docs.gitlab.com/ee/user/project/merge_requests/code_quality.html."""

    __implements__ = IReporter
    name = "gitlab-codeclimate"
    extension = "json"

    _MSG_TYPES = {
        "info": "info",
        "convention": "minor",
        "refactor": "major",
        "warning": "major",
        "error": "critical",
        "fatal": "blocker",
    }

    def __init__(self, output=sys.stdout):
        BaseReporter.__init__(self, output)
        self.messages = []

    def handle_message(self, msg):
        """Manage message of different type and in the context of path."""
        self.messages.append({
            "description": html.escape(msg.msg_id + ': ' + msg.msg or "", quote=False),
            "severity": self._MSG_TYPES[msg.category],
            "location": {
                "path": msg.path,
                "lines": {
                    "begin": msg.line,
                }
            },
            "fingerprint": hashlib.md5(":".join([msg.path, str(msg.line), msg.msg_id]).encode()).hexdigest()
        })

    def display_messages(self, layout):
        """Launch layouts display."""
        print(json.dumps(self.messages, indent=4), file=self.out)

    def display_reports(self, layout):
        """Do nothing."""

    def _display(self, layout):
        """Do nothing."""



#
# Copyright 2020 Stephan Mueller
#
# Licensed under the MIT license

"""Setup for pylint_gitlab module."""

import json
import logging
import os
import subprocess

import setuptools

logger = logging.getLogger(__name__)


def get_install_requirements() -> list:
    """Retrieves list of packages from Pipfile.lock required for installation.

    Returns: List of packages
    """
    with open("Pipfile.lock") as file:
        pipfile = file.read()

    packages = list()
    for name, _ in json.loads(pipfile)["default"].items():
        packages.append(name)
    return packages


def long_description() -> str:
    """Reads README.md

    Returns: Content of ``README.md``
    """
    with open("README.md", "r") as file:
        return file.read()


def version() -> str:
    """Tries to detect version based on selected strategy.

    Returns: Project version
    """

    version_strategy = os.getenv("VERSION_STRATEGY", "GIT_REF_NAME")

    if version_strategy == "GIT_COMMIT_SHA":

        if os.getenv("CI_COMMIT_SHA", "") != "":
            return os.getenv("CI_COMMIT_SHA")

        process = subprocess.run(["git", "rev-parse", "--quiet", "HEAD"], capture_output=True, check=True)
        commit_sha = process.stdout.decode().strip()
        if commit_sha is not None and commit_sha != "":
            return "0.0.0.commit" + commit_sha

    elif version_strategy == "GIT_REF_NAME":

        if os.getenv("CI_COMMIT_REF_NAME", "") != "":
            return os.getenv("CI_COMMIT_REF_NAME")

        process = subprocess.run(["git", "symbolic-ref", "--quiet", "HEAD"], capture_output=True, check=True)
        branch = process.stdout.decode().strip().replace("refs/heads/", "", 1)
        if branch is not None and branch != "":
            return branch

    raise ValueError("Version could not be detected.")


setuptools.setup(
    name="pylint-gitlab",
    version=version(),
    author="Stephan Müller",
    author_email="mail@stephanmueller.eu",
    license='MIT',
    description="This project provides pylint formatters for a nice integration with GitLab CI.",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    url="https://gitlab.com/smueller18/pylint-gitlab",
    packages=setuptools.find_packages(),
    install_requires=get_install_requirements(),
    package_data={
        "": ["templates/report.html.j2"]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    project_urls={
        'Documentation': 'https://smueller18.gitlab.com/pylint-gitlab/',
        'Source': 'https://gitlab.com/smueller18/pylint-gitlab',
        'Tracker': 'https://gitlab.com/smueller18/pylint-gitlab/issues',
    },
)
